﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Windows;
using Telerik.Windows.Controls.Map;

namespace MapGeometryDemo
{
    public class Group : BindableObject
    {
        private static readonly ISpatialReference SpatialReference = new SpatialReference();
        private static readonly Location BasePosition = new Location(0,0);

        private readonly IEnumerable<Thing> _things;
        private double _direction;
        private double _height;
        private double _width;
        private Location _location;

        public Group(IEnumerable<Thing> things)
        {
            _things = things;
            _location = new Location();
            foreach (var thing in _things)
            {
                thing.PropertyChanged += ThingOnPropertyChanged;
            }
            Update();
        }

        public Location Location
        {
            get { return _location; }
            set
            {
                if (_location == value)
                    return;
                _location = value;
                RaisePropertyChanged(() => Location);
            }
        }

        public double Width
        {
            get { return _width; }
            private set
            {
                if (_width == value)
                    return;
                _width = value;
                RaisePropertyChanged(() => Width);
            }
        }

        public double Height
        {
            get { return _height; }
            private set
            {
                if (_height == value)
                    return;
                _height = value;
                RaisePropertyChanged(() => Height);
            }
        }

        public double Direction
        {
            get { return _direction; }
            private set
            {
                if (_direction == value)
                    return;
                _direction = value;
                RaisePropertyChanged(() => Direction);
            }
        }

        private void ThingOnPropertyChanged(object sender, PropertyChangedEventArgs propertyChangedEventArgs)
        {
            Update();
        }

        private void Update()
        {
            var averageHeading = CalculateMeanAngle(_things.Select(thing => thing.Direction));
            var rotatedPoints = _things.Select(thing => RotatePointAroundOriginNedClockwise(ConvertLatLongToNorthEastRelative(thing.Location, BasePosition), -averageHeading)).ToList();
            var top = rotatedPoints.Select(point => point.Y).Max();
            var bottom = rotatedPoints.Select(point => point.Y).Min();
            var left = rotatedPoints.Select(point => point.X).Min();
            var right = rotatedPoints.Select(point => point.X).Max();

            Direction = averageHeading;
            var realWidth = right - left;
            Width = Math.Max(realWidth, 0);
            var realHeight = top - bottom;
            Height = Math.Max(realHeight, 0);

            var rotatedCenter = new Point(left + (realWidth / 2), bottom + (realHeight / 2));
            var center = RotatePointAroundOriginNedClockwise(rotatedCenter, averageHeading);
            Location = ConvertNorthEastRelativeToLatLong(center, BasePosition);
        }

        public static double CalculateMeanAngle(IEnumerable<double> angles)
        {
            return Math.Atan2(angles.Average(angle => Math.Sin(angle)), angles.Average(angle => Math.Cos(angle)));
        }

        public static Point RotatePointAroundOriginNedClockwise(Point point, double angleInRadian)
        {
            var sinAngle = Math.Sin(angleInRadian);
            var cosAngle = Math.Cos(angleInRadian);
            var x = point.X * cosAngle + point.Y * sinAngle;
            var y = -point.X * sinAngle + point.Y * cosAngle;

            return new Point(x, y);
        }

        public Location ConvertNorthEastRelativeToLatLong(Point point, Location basePosition)
        {
            var north = point.Y;
            var east = point.X;
            var heightInKm = north / 1000;
            var widthInKm = east / 1000;
            var sizeInDegree = SpatialReference.GetSizeInReferenceUnits(basePosition, new Size(Math.Abs(widthInKm), Math.Abs(heightInKm)));
            return new Location(basePosition.Latitude + Math.Sign(heightInKm) * sizeInDegree.Height,
                                basePosition.Longitude + Math.Sign(widthInKm) * sizeInDegree.Width);
        }

        public Point ConvertLatLongToNorthEastRelative(Location newPosition, Location basePosition)
        {
            var heightInDegrees = newPosition.Latitude - basePosition.Latitude;
            var widthInDegrees = newPosition.Longitude - basePosition.Longitude;
            var sizeInKm = SpatialReference.GetSizeInKilometers(basePosition, new Size(Math.Abs(widthInDegrees), Math.Abs(heightInDegrees)));
            return new Point(Math.Sign(widthInDegrees) * sizeInKm.Width * 1000, Math.Sign(heightInDegrees) * sizeInKm.Height * 1000);
        }
    }
}
