﻿using System;
using System.Windows;
using Telerik.Windows.Controls;
using Telerik.Windows.Controls.Map;

namespace MapGeometryDemo
{
    public class GroupView : MapGeometryView
    {
        public static readonly DependencyProperty CenterProperty = DependencyProperty.Register("Center", typeof(Location), typeof(GroupView), new PropertyMetadata(Location.Empty, PropertyChangedHandler));
        public static readonly DependencyProperty OrientationInRadProperty = DependencyProperty.Register("OrientationInRad", typeof(double), typeof(GroupView), new PropertyMetadata(0.0, PropertyChangedHandler));
        public static readonly DependencyProperty WidthInMProperty = DependencyProperty.Register("WidthInM", typeof(double), typeof(GroupView), new PropertyMetadata(200.0, PropertyChangedHandler));
        public static readonly DependencyProperty HeightInMProperty = DependencyProperty.Register("HeightInM", typeof(double), typeof(GroupView), new PropertyMetadata(1000.0, PropertyChangedHandler));
        public static readonly DependencyProperty PaddingProperty = DependencyProperty.Register("Padding", typeof(double), typeof(GroupView), new PropertyMetadata(0.0, PropertyChangedHandler));
        public static readonly DependencyProperty ParentRadMapProperty = DependencyProperty.Register("ParentRadMap", typeof(RadMap), typeof(GroupView), new PropertyMetadata(null, ParentRadMapChangedHandler));

        public Location Center
        {
            get { return (Location)GetValue(CenterProperty); }
            set { SetValue(CenterProperty, value); }
        }

        public double OrientationInRad
        {
            get { return (double)GetValue(OrientationInRadProperty); }
            set { SetValue(OrientationInRadProperty, value); }
        }

        public double WidthInM
        {
            get { return (double)GetValue(WidthInMProperty); }
            set { SetValue(WidthInMProperty, value); }
        }

        public double HeightInM
        {
            get { return (double)GetValue(HeightInMProperty); }
            set { SetValue(HeightInMProperty, value); }
        }

        public double Padding
        {
            get { return (double)GetValue(PaddingProperty); }
            set { SetValue(PaddingProperty, value); }
        }

        public RadMap ParentRadMap
        {
            get { return (RadMap)GetValue(ParentRadMapProperty); }
            set { SetValue(ParentRadMapProperty, value); }
        }

        private static void ParentRadMapChangedHandler(DependencyObject source, DependencyPropertyChangedEventArgs e)
        {
            var geometryView = source as GroupView;
            if (geometryView != null)
            {
                var oldParentRadMap = e.OldValue as RadMap;
                if (oldParentRadMap != null)
                {
                    geometryView.UnbindOfParentRadMap(oldParentRadMap);
                }
                var newParentRadMap = e.NewValue as RadMap;
                if (newParentRadMap != null)
                {
                    geometryView.BindToParentRadMap(newParentRadMap);
                }
                geometryView.UpdateFigure();
            }
        }

        private void BindToParentRadMap(RadMap parentRadMap)
        {
            parentRadMap.ZoomingFinished += ParentRadMapOnZoomingFinished;
            parentRadMap.ZoomChanged += ParentRadMapOnZoomChanged;
        }

        private void UnbindOfParentRadMap(RadMap parentRadMap)
        {
            parentRadMap.ZoomChanged -= ParentRadMapOnZoomChanged;
            parentRadMap.ZoomingFinished -= ParentRadMapOnZoomingFinished;
        }

        private void ParentRadMapOnZoomingFinished(object sender, RoutedEventArgs routedEventArgs)
        {
            UpdateFigure();
        }

        private void ParentRadMapOnZoomChanged(object sender, EventArgs eventArgs)
        {
            UpdateFigure();
        }

        private readonly ISpatialReference _spatialReference = new SpatialReference();

        public GroupView()
        {
            GeometryData = new PathGeometryData();
        }

        private static void PropertyChangedHandler(DependencyObject source, DependencyPropertyChangedEventArgs e)
        {
            var geometryView = source as GroupView;
            if (geometryView != null)
            {
                geometryView.UpdateFigure();
            }
        }

        private void UpdateFigure()
        {
            UpdateGeometry(GeometryData as PathGeometryData);
        }

        private void UpdateGeometry(PathGeometryData pathGeometryData)
        {
            pathGeometryData.BeginInit();
            pathGeometryData.Figures.Clear();

            if (!Center.Equals(Location.Empty) && ParentRadMap != null)
            {
                var halfHeight = HeightInM / 2;
                var halfWidth = WidthInM / 2;

                var lowerLeft = Group.RotatePointAroundOriginNedClockwise(new Point(-halfWidth, -halfHeight), OrientationInRad);
                var lowerRight = Group.RotatePointAroundOriginNedClockwise(new Point(halfWidth, -halfHeight), OrientationInRad);
                var midLeft = Group.RotatePointAroundOriginNedClockwise(new Point(-halfWidth, halfHeight), OrientationInRad);
                var midRight = Group.RotatePointAroundOriginNedClockwise(new Point(halfWidth, halfHeight), OrientationInRad);
                var tip = Group.RotatePointAroundOriginNedClockwise(new Point(0, halfHeight + CalculateTipHeight()), OrientationInRad);

                var lowerLeftPixelOffset = Group.RotatePointAroundOriginNedClockwise(new Point(-Padding, -Padding), OrientationInRad);
                var lowerRightPixelOffset = Group.RotatePointAroundOriginNedClockwise(new Point(Padding, -Padding), OrientationInRad);
                var midLeftPixelOffset = Group.RotatePointAroundOriginNedClockwise(new Point(-Padding, Padding), OrientationInRad);
                var midRightPixelOffset = Group.RotatePointAroundOriginNedClockwise(new Point(Padding, Padding), OrientationInRad);
                var tipPixelOffset = Group.RotatePointAroundOriginNedClockwise(new Point(0, Padding*2), OrientationInRad);

                var figure = new PathFigureData
                {
                    IsClosed = true,
                    IsFilled = true,
                    StartPoint = CalculateLocationAtPixelOffset(GetLocation(lowerLeft, Center), lowerLeftPixelOffset),
                    Segments =
                    {
                        new LineSegmentData{Point = CalculateLocationAtPixelOffset(GetLocation(lowerRight, Center), lowerRightPixelOffset)},
                        new LineSegmentData{Point = CalculateLocationAtPixelOffset(GetLocation(midRight, Center), midRightPixelOffset)},
                        new LineSegmentData{Point = CalculateLocationAtPixelOffset(GetLocation(tip, Center), tipPixelOffset)},
                        new LineSegmentData{Point = CalculateLocationAtPixelOffset(GetLocation(midLeft, Center), midLeftPixelOffset)}
                    }
                };
                pathGeometryData.Figures.Add(figure);
            }

            pathGeometryData.EndInit();
        }

        private Location GetLocation(Point point, Location baseLocation)
        {
            var latLongSize = _spatialReference.GetSizeInReferenceUnits(baseLocation, new Size(Math.Abs(point.X / 1000), Math.Abs(point.Y / 1000)));
            return new Location(baseLocation.Latitude + (latLongSize.Height * Math.Sign(point.Y)),
                                baseLocation.Longitude + (latLongSize.Width * Math.Sign(point.X)));
        }

        private double CalculateTipHeight()
        {
            return 0.433012701 * WidthInM;
        }

        private Location CalculateLocationAtPixelOffset(Location baseLocation, Point pixelOffset)
        {
            var baseLocationInPixel = baseLocation.GetPoint(ParentRadMap);
            var offsetLocationInPixel = new Point(baseLocationInPixel.X + pixelOffset.X, baseLocationInPixel.Y - pixelOffset.Y);
            return Location.GetCoordinates(ParentRadMap, offsetLocationInPixel);
        }
    }
}
